@@ -1,5 +1,5 @@
 /*
- * growisofs 5.12 by Andy Polyakov <appro@fy.chalmers.se>.
+ * growisofs 5.18 by Andy Polyakov <appro@fy.chalmers.se>.
  *
  * Use-it-on-your-own-risk, GPL bless...
  *
@@ -199,12 +199,94 @@
  * - report I/O error if input stream was less than 64K;
  * - -M /dev/cdrom=/dev/zero didn't relocate the lead-out in DVD-RW
  *   Restricted Overwrite;
+ * 5.13:
+ * - workarounds for Panasonic/MATSUSHITA DVD-RAM LF-D310;
+ * - Solaris: media load upon start-up;
+ * 5.14:
+ * - LG GSA-4040B failed to auto-format DVD+RW blanks;
+ * - '| growisofs -Z /dev/cdrom=/dev/fd/0' failed with "already carries
+ *   isofs" even when running interactively, so I check on /dev/tty
+ *   instead of isatty(0);
+ * - error output was confusing when overburn condition was raised in
+ *   dry-run mode;
+ * - more sane default drain buffer size to minimize system load when
+ *   unit fails to return usable buffer utilization statistics under
+ *   "LONG WRITE IN PROGRESS" condition;
+ * - progress indicator process was orphaned if -Z /dev/cdrom=file.iso
+ *   terminated prematurely;
+ * - -overburn -Z /dev/cdrom=file.iso printed two "ignored" messages;
+ * - Solaris: use large-file API in setup_fds;
+ * - HP-UX: HP-UX support is contributed by HP;
+ * - block signals in the beginning of recording, formally it shouldn't
+ *   be necessary, but is apparently needed for some units (is it?);
+ * - prepare code for -speed even in DVD+ context, need a test-case...
+ * - TEAC DV-W50D and Lite-On LDW-811S failed to set recording velocity,
+ *   deploy GET PERFORMANCE/SET STREAMING commands;
+ * - Lite-On LDW-811S returns 0s in Write Speed descriptors in page 2A,
+ *   this would cause a problem if DVD+ speed control was implemented;
+ * 5.15:
+ * - confusing output when DAO mode is manually engaged and DVD-RW media
+ *   is minimally blanked;
+ * - complement -use-the-force-luke=dao[:size] to arrange for piping
+ *   non-iso images in DAO mode (size is to be expressed in 2KB chunks);
+ * - Pioneer DVR-x06 apparently needs larger timeout to avoid "the LUN
+ *   appears to be stuck" message in the beginning of DAO recording;
+ * - HP-UX: fix-up umount code;
+ * - HP-UX: make sure user doesn't pass /dev/rscsi/cXtYlZ, they should
+ *   stick to /dev/rdsk/cXtYdZ;
+ * - implement -use-the-force-luke=seek:N -Z /dev/dvd=image to arrange
+ *   for 'builtin_dd if=image of=/dev/dvd obs=32k seek=N/16' (note that
+ *   N is expected to be expressed in 2KB chunks);
+ * - skip overwrite check for blank media to avoid read errors at start,
+ *   which reportedly may cause bus reset in some configurations;
+ * - make get_mmc_profile load media, explicit media load used to be on
+ *   per platform basis, while it doesn't have to;
+ * - postpone handle_events till after dry-run checkpoint;
+ * - error reporting revised;
+ * - Optorite seems to insist on resuming suspended DVD+RW format, at
+ *   least it's apparently the only way to get *reliable* results
+ *   (formally this contradicts specification, which explicitly states
+ *   that format is to be resumed automatically and transparently);
+ * - FreeBSD: FreeBSD 5-CURRENT since 2003-08-24, including 5.2 fails
+ *   to pull sense data automatically, at least for ATAPI transport,
+ *   so I reach for it myself (it's apparently a kernel bug, which
+ *   eventually will be fixed, but I keep the workaround code just in
+ *   case);
+ * - -speed option in DVD+ context is enabled, upon release tested with
+ *   Plextor PX-708A;
+ * - make builtin_dd print amount of transferred data, together with
+ *   -use-the-force-luke=seek:N it's easier to maintain "tar-formatted"
+ *   rewritable media;
+ * 5.16:
+ * - brown-bag bug in "LONG WRITE IN PROGRESS" code path;
+ * 5.17:
+ * - Linux: fix for /proc/sys/dev/cdrom/check_media set to 1;
+ * - HP-UX: INQUIRY buffer is required to be 128 bytes. Well, "required"
+ *   is wrong word in this context, as it's apparently a kernel bug
+ *   addressed in PHKL_30038 (HPUX 11.11) and PHKL_30039 (HPUX 11.23).
+ *   This "change" affects all dvd+rw-tools, but I don't bump their
+ *   version numbers for this, as it's an "ugly" workaround for an
+ *   *external* problem;
+ * - switch to GET PERFORMANCE even for current write speed (most
+ *   notably required for NEC and derivatives);
+ * - the above change required adaptations for Pioneer and LG units,
+ *   which don't/fail to provide current write speed through GET
+ *   PERFORMANCE despite the fact that the command is mandatory;
+ * - HP-UX: retain root privileges in setup_fds, SIOC_IO requires them;
+ * - fix for COMMAND SEQUENCE ERROR in the beginning of DVD-recording;
+ * - drop privileges prior mkisofs -version;
+ * 5.18:
+ * - refuse to run if ${SUDO_COMMAND} is set;
+ * - minimize amount of compiler warnings on 64-bit platforms;
+ * - skip count-down if no_tty_check is set;
+ * - -use-the-force-luke=tracksize:size option by suggestion from K3b;
+ * - Linux: fix for "Bad file descriptor" with DVD+RW kernel patch;
  */
 #define PRINT_VERSION(cmd)	do {			\
     char *s=strrchr((cmd),'/');				\
     s ? s++ : (s=(cmd));				\
     printf ("* %.*sgrowisofs by <appro@fy.chalmers.se>,"\
-	    " version 5.12,\n",(int)(s-(cmd)),(cmd));	\
+	    " version 5.18,\n",(int)(s-(cmd)),(cmd));	\
 } while (0)
 
 #define _LARGEFILE_SOURCE 
@@ -260,7 +342,7 @@ typedef ssize_t (*pwrite64_t)(int,const void *,size_t,off64_t);
  */
 int        get_mmc_profile	(void *fd);
 int        plusminus_r_C_parm	(void *fd,char *C_parm);
-pwrite64_t poor_mans_setup	(void *fd,int profile,off64_t leadout);
+pwrite64_t poor_mans_setup	(void *fd,off64_t leadout);
 /*
  * These never terminate the program.
  * Pay attention to the return value.
@@ -319,12 +401,12 @@ static int    in_fd=-1,out_fd=-1;
 #define INVALID_HANDLE ((void *)-1)
 #endif
 static void   *ioctl_handle=INVALID_HANDLE;
-#define ioctl_fd ((int)ioctl_handle)
+#define ioctl_fd ((long)ioctl_handle)
 
-static int	poor_man=0, zero_image=0, quiet=1,
-		overburn=0, no_tty_check=0, dry_run=0;
+static int	poor_man=-1, zero_image=0, quiet=1,
+		overburn=0, no_tty_check=0, dry_run=0, dao_size=0;
 
-int    dvd_compat=0, test_write=0, no_reload=0, mmc_profile=-1;
+int    dvd_compat=0, test_write=0, no_reload=0, mmc_profile=0;
 double speed_factor=0.0;
 char  *ioctl_device;
 int    _argc;
@@ -340,7 +422,7 @@ char *find_raw_device(struct stat64 *sb)
 { int   i,rawctl,dev_major,dev_minor;
   char  *ret=NULL;
   struct raw_config_request req;
-  static char rawdevname[16] = "";
+  static char rawdevname[24] = "";
 
     if (!S_ISBLK(sb->st_mode)) return NULL;
 
@@ -391,7 +473,7 @@ char *setup_fds (char *device)
 #endif
 
     if ((in_fd = open64 (device,O_RDONLY)) < 0)
-	if (errno != ENOMEDIUM ||
+	if (!(errno == ENOMEDIUM || errno == EMEDIUMTYPE) ||
 	    (in_fd = open64 (device,O_RDONLY|O_NONBLOCK)) < 0)
 	    fprintf (stderr,":-( unable to open64(\"%s\",O_RDONLY): ",device),
 	    perror (NULL), exit (FATAL_START(errno));
@@ -429,7 +511,7 @@ char *setup_fds (char *device)
 	return device;
     }
 
-    ioctl_handle=(void *)dup(in_fd);
+    ioctl_handle=(void *)(long)dup(in_fd);
 
     /*
      * get_mmc_profile terminates the program if ioctl_handle is
@@ -444,7 +526,7 @@ char *setup_fds (char *device)
 	perror (NULL), exit (FATAL_START(errno));
     }
 
-    switch (mmc_profile)
+    switch (mmc_profile&0xFFFF)
     {	case 0x11:	/* DVD-R			*/
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
@@ -455,12 +537,12 @@ char *setup_fds (char *device)
 	    setuid (uid);	/* drop all privileges	*/
 	    return ioctl_device=device;
 	case 0x1A:	/* DVD+RW			*/
-	    if (poor_man) goto open_poor_man;
+	    if (poor_man>0) goto open_poor_man;
 	case 0x12:	/* DVD-RAM			*/
 	    break;
 	default:
-	    fprintf (stderr,":-( %s: unsupported MMC profile %X\n",
-			    device,mmc_profile);
+	    fprintf (stderr,":-( %s: media is not recognized as "
+			    "recordable DVD: %X\n",device,mmc_profile);
 	    exit (FATAL_START(EMEDIUMTYPE));
     }
 
@@ -483,8 +565,8 @@ char *setup_fds (char *device)
 	device=odevice;
 	goto opened_rw;
     }
-    if (mmc_profile == 0x12)	goto open_rw;	/* DVD-RAM */
-    else			goto open_poor_man;
+    if ((mmc_profile&0xFFFF) == 0x12)	goto open_rw;	/* DVD-RAM */
+    else				goto open_poor_man;
 }
 
 #elif defined(__OpenBSD__) || defined(__NetBSD__)
@@ -534,7 +616,7 @@ char *setup_fds (char *device)
      * this means that if installed set-root-uid, growisofs grants
      * access to DVD burner(s), but not to any other device.
      */
-    ioctl_handle = (void *)open (device,O_RDWR); /* O_RDWR is a must for SCIOCCOMMAND */
+    ioctl_handle = (void *)(long)open (device,O_RDWR); /* O_RDWR is a must for SCIOCCOMMAND */
     if (ioctl_handle == INVALID_HANDLE)
     {	fprintf (stderr,":-( unable to open(\"%s\",O_RDWR): ",device),
 	perror (NULL);
@@ -550,7 +632,7 @@ char *setup_fds (char *device)
 	exit(FATAL_START(EPERM));
 
     mmc_profile = get_mmc_profile (ioctl_handle);
-    switch (mmc_profile)
+    switch (mmc_profile&0xFFFF)
     {	case 0x11:	/* DVD-R			*/
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
@@ -564,8 +646,8 @@ char *setup_fds (char *device)
 	    out_fd = dup(ioctl_fd);
 	    goto opened_rw;
 	default:
-	    fprintf (stderr,":-( %s: unsupported MMC profile %X\n",
-			    device,mmc_profile);
+	    fprintf (stderr,":-( %s: media is not recognized as "
+			    "recordable DVD: %X\n",device,mmc_profile);
 	    exit (FATAL_START(EMEDIUMTYPE));
     }
     /* not actually reached */
@@ -659,7 +741,7 @@ char *setup_fds (char *device)
     ioctl_handle = (void *)cam;
 
     mmc_profile = get_mmc_profile (ioctl_handle);
-    switch (mmc_profile)
+    switch (mmc_profile&0xFFFF)
     {	case 0x11:	/* DVD-R			*/
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
@@ -672,8 +754,8 @@ char *setup_fds (char *device)
 	case 0x12:	/* DVD-RAM			*/
 	    break;
 	default:
-	    fprintf (stderr,":-( %s: unsupported MMC profile %X\n",
-			    device,mmc_profile);
+	    fprintf (stderr,":-( %s: media is not recognized as "
+			    "recordable DVD: %X\n",device,mmc_profile);
 	    exit (FATAL_START(EMEDIUMTYPE));
     }
     setuid(uid);
@@ -683,10 +765,11 @@ char *setup_fds (char *device)
 #elif defined(__sun) || defined(sun)
 
 #include <volmgt.h>
+#include <sys/cdio.h>
 
 char *setup_fds (char *device)
 { uid_t uid=getuid();
-  struct stat sb,sc;
+  struct stat64 sb,sc;
   int v;
 
     if ((v=volmgt_running()))
@@ -707,13 +790,13 @@ char *setup_fds (char *device)
     /*
      * We might be entering as euid=root!
      */
-    if ((in_fd = open (device,O_RDONLY)) < 0)
+    if ((in_fd = open64 (device,O_RDONLY)) < 0)
 	if (errno != ENXIO ||
-	    (in_fd = open (device,O_RDONLY|O_NONBLOCK)) < 0)
+	    (in_fd = open64 (device,O_RDONLY|O_NONBLOCK)) < 0)
 	    fprintf (stderr,":-( unable to open(\"%s\",O_RDONLY): ",device),
 	    perror (NULL), exit (FATAL_START(errno));
 
-    if (fstat (in_fd,&sb) < 0)
+    if (fstat64 (in_fd,&sb) < 0)
 	fprintf (stderr,":-( unable to stat(\"%s\"): ",device),
 	perror (NULL), exit (FATAL_START(errno));
 
@@ -721,19 +804,19 @@ char *setup_fds (char *device)
     { char *s;
 	if (S_ISBLK(sb.st_mode) && (s=strstr (device,"/dsk/")))
 	{   fprintf (stderr,":-) you most likely want to use %.*s/r%s instead!\n",
-			    s-device,device,s+1);
+			    (int)(s-device),device,s+1);
 	    if (isatty(0) && !dry_run) poll(NULL,0,5000);
 	}
 	setuid(uid);	/* drop all privileges	*/
 	close (in_fd);	/* reopen as mortal	*/
-	if ((in_fd = open (device,O_RDONLY)) < 0)
+	if ((in_fd = open64 (device,O_RDONLY)) < 0)
 	    fprintf (stderr,":-( unable to reopen(\"%s\",O_RDONLY): ",device),
 	    perror (NULL), exit (FATAL_START(errno));
       open_rw:
-	if ((out_fd = open (device,O_RDWR)) < 0)
+	if ((out_fd = open64 (device,O_RDWR)) < 0)
 	    fprintf (stderr,":-( unable to open(\"%s\",O_RDWR): ",device),
 	    perror (NULL), exit (FATAL_START(errno));
-	if (fstat(out_fd,&sc) < 0)
+	if (fstat64(out_fd,&sc) < 0)
 	    fprintf (stderr,":-( unable to stat(\"%s\"): ",device),
 	    perror (NULL), exit (FATAL_START(errno));
 	if (sb.st_dev!=sc.st_dev || sb.st_ino!=sc.st_ino)
@@ -753,10 +836,15 @@ char *setup_fds (char *device)
      * this means that if installed set-root-uid, growisofs grants
      * access to DVD burner(s), but not to any other device.
      */
-    ioctl_handle = (void *)dup (in_fd);
+    ioctl_handle = (void *)(long)dup (in_fd);
+
+#if 0
+    if (ioctl(ioctl_handle,CDROMSTART)<0 && errno==ENXIO)
+	media_load(ioctl_handle);
+#endif
 
     mmc_profile = get_mmc_profile (ioctl_handle);
-    switch (mmc_profile)
+    switch (mmc_profile&0xFFFF)
     {	case 0x11:	/* DVD-R			*/
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
@@ -764,7 +852,7 @@ char *setup_fds (char *device)
 	case 0x1A:	/* DVD+RW			*/
 	    poor_man = 1;
 	    out_fd = dup(ioctl_fd);
-#if 0	    /* 'man uscsi' maintains that root priveleges are required upon
+#if 0	    /* 'man uscsi' maintains that root privileges are required upon
 	     * issue of USCSI ioctl, we therefore can't drop them...
 	     */
 	    setuid (uid);	/* drop all privileges	*/
@@ -773,22 +861,234 @@ char *setup_fds (char *device)
 	case 0x12:	/* DVD-RAM			*/
 	    break;
 	default:
-	    fprintf (stderr,":-( %s: unsupported MMC profile %X\n",
-			    device,mmc_profile);
+	    fprintf (stderr,":-( %s: media is not recognized as "
+			    "recordable DVD: %X\n",device,mmc_profile);
 	    exit (FATAL_START(EMEDIUMTYPE));
     }
     setuid(uid);
     goto open_rw;
 }
 
+#elif defined(__hpux)
+
+#ifdef SCTL_MAJOR
+#  define SCTL_SANITY_CHECK SCTL_MAJOR-1
+#  if SCTL_SANITY_CHECK<=0
+#    undef SCTL_MAJOR
+#  endif
+#  undef SCTL_SANITY_CHECK
+#endif
+
+#ifndef SCTL_MAJOR
+#error "SCTL_MAJOR is undefined or not sane."
+#endif
+
+#include <sys/mknod.h>
+#include <sys/diskio.h>
+
+#define seteuid(x)	setreuid((uid_t)-1,x)
+
+#if 1
+#define CANNOT_PASS_DEV_FD_N_TO_MKISOFS
+#elif 0
+--- ./multi.c.orig	Wed Dec 25 15:15:24 2002
++++ ./multi.c	Tue Nov 11 17:12:27 2003
+@@ -1067,3 +1067,13 @@
+ open_merge_image(path)
+ 	char	*path;
+ {
++	int fd;
++
++	if (sscanf (path,"/dev/fd/%u",&fd) == 1) {
++		int fdd = dup(fd);	/* validate file descriptor */
++		if (fdd < 0) return -1;
++		close (fdd);
++		in_image = fdopen (fd,"rb");
++		return in_image ? 0 : -1;
++	}
++
+#endif
+
+#ifdef CANNOT_PASS_DEV_FD_N_TO_MKISOFS
+char *get_M_parm (int fd, char *device)
+{ struct stat sb;
+  dev_t m;
+  char *ret=device;
+  static char ctl[16];
+
+    if (fstat (fd,&sb)==0 && S_ISCHR(sb.st_mode))
+    {	m = minor (sb.st_rdev);
+	sprintf (ctl,"%d,%d,%d",(m>>16)&0xFF,(m>>12)&0xF,(m>>8)&0xF);
+	ret = ctl;
+    }
+
+  return ret;
+}
+#endif
+
+char *setup_fds (char *device)
+{ uid_t uid=getuid();
+  struct stat64 sb,sc;
+  dev_t  m;
+  static char rscsi [32];
+  disk_describe_type ddt;
+
+    /*
+     * We might be entering as euid=root!
+     */
+    if ((in_fd = open64 (device,O_RDONLY)) < 0)
+	fprintf (stderr,":-( unable to open(\"%s\",O_RDONLY): ",device),
+	perror (NULL), exit (FATAL_START(errno));
+
+    if (fstat64 (in_fd,&sb) < 0)
+	fprintf (stderr,":-( unable to stat(\"%s\"): ",device),
+	perror (NULL), exit (FATAL_START(errno));
+
+    if (!S_ISCHR(sb.st_mode))
+    { char *s;
+	if (S_ISBLK(sb.st_mode) && (s=strstr (device,"/dsk/")))
+	{   fprintf (stderr,":-) you most likely want to use %.*s/r%s instead!\n",
+			    (int)(s-device),device,s+1);
+	    if (isatty(0) && !dry_run) poll(NULL,0,5000);
+	}
+	setuid(uid);	/* drop all privileges	*/
+	close (in_fd);	/* reopen as mortal	*/
+	if ((in_fd = open64 (device,O_RDONLY)) < 0)
+	    fprintf (stderr,":-( unable to reopen(\"%s\",O_RDONLY): ",device),
+	    perror (NULL), exit (FATAL_START(errno));
+      open_rw:
+	if ((out_fd = open64 (device,O_RDWR)) < 0)
+	    fprintf (stderr,":-( unable to open(\"%s\",O_RDWR): ",device),
+	    perror (NULL), exit (FATAL_START(errno));
+	if (fstat64(out_fd,&sc) < 0)
+	    fprintf (stderr,":-( unable to stat(\"%s\"): ",device),
+	    perror (NULL), exit (FATAL_START(errno));
+	if (sb.st_dev!=sc.st_dev || sb.st_ino!=sc.st_ino)
+	    fprintf (stderr,":-( %s: race condition detected!\n",device),
+	    exit(FATAL_START(EPERM));
+	poor_man = 0;
+	close (ioctl_fd);
+	ioctl_handle = INVALID_HANDLE;
+	setuid (uid);	/* drop all privileges	*/
+	return device;
+    }
+
+    /*
+     * Still as euid=root! But note that get_mmc_profile makes sure it's
+     * an MMC device, as it terminates the program if the unit doesn't
+     * reply to GET CONFIGURATON. In combination with following switch
+     * this means that if installed set-root-uid, growisofs grants
+     * access to DVD burner(s), but not to any other device.
+     */
+
+    m=minor(sb.st_rdev);
+
+    /* Make sure user isn't using /dev/rscsi/cXtYlZ... */
+#if 1
+    if (ioctl (in_fd,DIOC_DESCRIBE,&ddt) != 0)
+#else
+    if (major(sb.st_rdev) == SCTL_MAJOR)
+#endif
+	fprintf (stderr,":-( stick to /dev/rdsk/c%xt%x%c%x!\n",
+			(m>>16)&0xFF,(m>>12)&0xF,'d',(m>>8)&0xF),
+	exit(FATAL_START(EINVAL));
+
+    /*
+     * Even though /dev/rdsk/cXtYdZ accepts SIOC_IO as well, we have to
+     * use /dev/rscsi/cXtYlZ for pass-through access in order to avoid
+     * command replay by upper "class" driver...
+     */
+    sprintf (rscsi,"/dev/rscsi/c%xt%x%c%x",
+		    (m>>16)&0xFF,(m>>12)&0xF,'l',(m>>8)&0xF);
+    ioctl_handle = (void *)open64 (rscsi,O_RDONLY);
+    if (ioctl_handle == INVALID_HANDLE)
+    {	fprintf (stderr,":-( unable to open(\"%s\",O_RDONLY): ",rscsi),
+	perror (NULL);
+	if (errno == ENOENT)
+	    fprintf (stderr,":-! consider "
+			    "'mknod %s c %d 0x%06x; chmod 0600 %s'\n",
+			    rscsi,SCTL_MAJOR,(m&0xFFFF00)|2,rscsi);
+	exit (FATAL_START(errno));
+    }
+    if (fstat64 (ioctl_fd,&sc) < 0)
+	fprintf (stderr,":-( unable to stat(\"%s\"): ",rscsi),
+	perror (NULL), exit (FATAL_START(errno));
+    /* Make sure we land on same SCSI ID... */
+    if ((m&0xFFFF00) != (minor(sc.st_rdev)&0xFFFF00))
+	fprintf (stderr,":-( SCSI ID mismatch: %06x!=%06x\n",
+			m,minor(sc.st_rdev)),
+	exit(FATAL_START(EPERM));
+
+    mmc_profile = get_mmc_profile (ioctl_handle);
+    switch (mmc_profile&0xFFFF)
+    {	case 0x11:	/* DVD-R			*/
+	case 0x13:	/* DVD-RW Restricted Overwrite	*/
+	case 0x14:	/* DVD-RW Sequential		*/
+	case 0x1B:	/* DVD+R			*/
+	case 0x1A:	/* DVD+RW			*/
+	case 0x12:	/* DVD-RAM			*/
+	    poor_man = 1;
+	    out_fd = dup(ioctl_fd);
+#if 0	    /* HP-UX requires root privileges upon SIOC_IO ioctl */
+	    setuid (uid);	/* drop all privileges	*/
+#endif
+	    return ioctl_device=rscsi;
+	default:
+	    fprintf (stderr,":-( %s: media is not recognized as "
+			    "recordable DVD: %X\n",device,mmc_profile);
+	    exit (FATAL_START(EMEDIUMTYPE));
+    }
+    setuid(uid);
+    goto open_rw;
+}
+
+/*
+ * PA-RISC HP-UX doesn't have /dev/zero:-(
+ */
+static fd_set _dev_zero;
+
+static int open64_zero(const char *pathname, int flags)
+{ int fd;
+    if (strcmp(pathname,"/dev/zero"))
+	return open64 (pathname,flags);
+    else if ((fd=open ("/dev/null",flags)) >= 0)
+	FD_SET (fd,&_dev_zero);
+  return fd;    
+}
+
+static ssize_t read_zero (int fd, void *buf, size_t count)
+{   if (!FD_ISSET(fd,&_dev_zero))	return read (fd,buf,count);
+    memset (buf,0,count);		return count;
+}
+
+static int close_zero (int fd)
+{ int ret=close (fd);
+    if (ret>=0 && FD_ISSET(fd,&_dev_zero))	FD_CLR (fd,&_dev_zero);
+  return ret;
+}
+
+static int dup2_zero (int oldfd, int newfd)
+{ int ret;
+    ret = dup2 (oldfd,newfd);
+    if (ret >= 0)
+    {	if (FD_ISSET(oldfd,&_dev_zero))	FD_SET (ret,&_dev_zero);
+	else				FD_CLR (ret,&_dev_zero);
+    }
+  return ret;
+}
+#define open64	open64_zero
+#define read	read_zero
+#define close	close_zero
+#define dup2	dup2_zero
+
 #else
 #error "Unsupported OS"
 #endif
 
 int setup_C_parm (char *C_parm,struct iso_primary_descriptor *descr)
-{ int next_session=-1;
+{ int next_session=-1,profile=mmc_profile&0xFFFF;
 
-    if (!poor_man || mmc_profile==0x1A || mmc_profile==0x13)
+    if (!poor_man || profile==0x1A || profile==0x13)
     {	next_session = from_733(descr->volume_space_size);
 	/* pad to closest 32K boundary */
 	next_session += 15;
@@ -796,7 +1096,7 @@ int setup_C_parm (char *C_parm,struct iso_primary_descriptor *descr)
 	next_session *= 16;
 	sprintf (C_parm,"16,%u",next_session);
     }
-    else if (mmc_profile==0x1B || mmc_profile==0x11 || mmc_profile==0x14)
+    else if (profile==0x1B || profile==0x11 || profile==0x14)
 	next_session=plusminus_r_C_parm (ioctl_handle,C_parm);
 
   return next_session;
@@ -806,8 +1106,8 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 { char		*s;
   int   	 n,off,fd;
   struct stat64  sb;
-  off64_t	 capacity=0,tracksize=0;
-  pid_t		 pid=(pid_t)-1;
+  off64_t	 capacity=0,tracksize=0,startoff=outoff;
+  pid_t		 pid=(pid_t)-1,ppid=getpid();
   volatile struct { time_t zero; off64_t current,final; } *progress;
 
     if ((fd=mkstemp (s=strdup("/tmp/dvd+rw.XXXXXX"))) < 0)
@@ -833,15 +1133,18 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 
     progress->zero=0;
     progress->current=outoff;
-    if (S_ISREG(sb.st_mode))
-    {	progress->final=outoff+(tracksize=sb.st_size);
+    if (dao_size || S_ISREG(sb.st_mode))
+    {	tracksize = dao_size ? (dao_size*CD_BLOCK) : sb.st_size;
+	progress->final=outoff+tracksize;
 	if (capacity && progress->final > capacity)
-	{   fprintf (stderr,":-( %s: not enough space available "
-			    "to accomodate this volume!\n",
-			    ioctl_device);
+	{   fprintf (stderr,":-( %s: %lld blocks are free, "
+			    "%lld to be written!\n",
+			    ioctl_device,
+			    (capacity-outoff)/2048,tracksize/2048);
 	    if (overburn)
 		fprintf (stderr,":-! ignoring...\n");
 	    else
+		close(infd), close(outfd),
 		exit (FATAL_START(ENOSPC));
 	}
     }
@@ -855,7 +1158,7 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 
 	close (infd);
 	close (outfd);
-	while (1)
+	while (kill (ppid,0)==0)
 	{   gettimeofday (&tv,NULL); slept =  tv.tv_usec/1e6+tv.tv_sec;
 	    poll (NULL,0,3333);
 	    gettimeofday (&tv,NULL); slept -= tv.tv_usec/1e6+tv.tv_sec;
@@ -877,7 +1180,7 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 				current,progress->final,0.0);
 	    fflush (stdout);
 	}
-	exit(0); /* not reached */
+	exit(0); /* not [normally] reached */
     }
 
     /* suck in first 64K and examine ISO9660 Primary Descriptor if present */
@@ -892,9 +1195,8 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 
 		if (!zero_image)
 	        {   if (tracksize==0)
-			tracksize=from_733(descr->volume_space_size)*CD_BLOCK;
-		    if (capacity)
-		    {	if (outoff+tracksize > capacity)
+		    {	tracksize=from_733(descr->volume_space_size)*CD_BLOCK;
+			if (capacity && (outoff+tracksize) > capacity)
 		    	{   fprintf (stderr,":-( %s: %lld blocks are free, "
 					    "%lld to be written\n",
 					    ioctl_device,
@@ -902,12 +1204,16 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 					    tracksize/2048);
 			    if (overburn)
 				fprintf (stderr,":-! ignoring...\n");
+			    else if (dry_run)
+				close(infd), close(outfd),
+				exit(FATAL_START(ENOSPC));
 			    else
 			    {	n = -1; errno = FATAL_START(ENOSPC);
 				goto out;
 			    }
 			}
 		    }
+		    /* else already checked for overburn condition */
 		}
 		else if (capacity > outoff)
 		{ int i=0;
@@ -937,8 +1243,8 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 		 * See commentary section in growisofs_mmc.cpp for
 		 * further details on poor_mans_setup
 		 */
-		pwrite64_method = poor_mans_setup (ioctl_handle,mmc_profile,
-							outoff+tracksize);
+		pwrite64_method = poor_mans_setup (ioctl_handle,
+						   outoff+tracksize);
 
 	    if (!progress->final)
 	    {	if (tracksize)	progress->final = outoff+tracksize;
@@ -1008,6 +1314,8 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 	}
 	outoff += DVD_BLOCK;
     }
+
+    printf ("builtin_dd: %lld*2KB out\n",(outoff-startoff)/2048);
 out:
     { int saved_errno=errno;
 
@@ -1078,15 +1386,22 @@ void pipe_mkisofs_up (char *mkisofs_argv[],int infd,int outfd,off64_t outoff)
 #define IVDs_SIZE	(sizeof(struct iso_primary_descriptor)*MAX_IVDs)
 
 int main (int argc, char *argv[])
-{ int  fd,imgfd=-1;
+{ int  imgfd=-1;
   char *in_device=NULL,*out_device=NULL,*in_image=NULL,*env;
   char dev_found;
   int  i,n,warn_for_isofs=0;
-  char **mkisofs_argv,C_parm[24],M_parm[16];
+  char **mkisofs_argv,C_parm[24],M_parm_[16],*M_parm=M_parm_;
   int  mkisofs_argc,growisofs_argc;
   int  next_session=-1,alleged_next_session=-1,new_size;
   struct iso_primary_descriptor *descr;
 
+    if (getenv ("SUDO_COMMAND"))
+    {	fprintf (stderr,":-( %s is being executed under sudo, "
+			"aborting!\n",argv[0]);
+	fprintf (stderr,"    See NOTES paragraph in growisofs "
+			"manual page for further details.\n");
+	exit(FATAL_START(EACCES));
+    }
     /*
      * This is a set-root-uid "entry point" for listed operations. User
      * can't trick this code to unmount arbitrary file system, as [s]he
@@ -1120,12 +1435,20 @@ int main (int argc, char *argv[])
     }
 
     /* mmap buffer so that we can use it with /dev/raw/rawN */
-    if ((fd=open ("/dev/zero",O_RDWR)) < 0)
-	perror (":-( unable to open \"/dev/zero\"?"),
-	exit(FATAL_START(errno));
+#if defined(MAP_ANONYMOUS) && !(defined(__sun) || defined(sun))
     the_buffer = mmap (NULL,2*DVD_BLOCK,PROT_READ|PROT_WRITE,
+			MAP_PRIVATE|MAP_ANONYMOUS,-1,0);
+#else
+    { int fd;
+
+	if ((fd=open ("/dev/zero",O_RDWR)) < 0)
+	    perror (":-( unable to open \"/dev/zero\"?"),
+	    exit(FATAL_START(errno));
+	the_buffer = mmap (NULL,2*DVD_BLOCK,PROT_READ|PROT_WRITE,
 			MAP_PRIVATE,fd,0);
-    close (fd);
+	close (fd);
+    }
+#endif
     if (the_buffer == MAP_FAILED)
 	fprintf (stderr,":-( unable to anonymously mmap %d: ",2*DVD_BLOCK),
 	perror (NULL), exit (FATAL_START(errno));
@@ -1170,7 +1493,7 @@ int main (int argc, char *argv[])
 		dev_found = 'Z';
 	    }
 	    else if (!strcmp(opt,"-poor-man"))
-	    {	poor_man = 1;
+	    {	if (poor_man<0) poor_man = 1;
 		continue;
 	    }
 	    else if (!strncmp(opt,"-speed",6))
@@ -1185,7 +1508,8 @@ int main (int argc, char *argv[])
 		continue;
 	    }
 	    else if (!strcmp(opt,"-dvd-compat"))
-	    {	poor_man = 1, dvd_compat++;
+	    {	if (poor_man<0) poor_man = 1;
+		dvd_compat++;
 		continue;
 	    }
 	    else if (!strcmp(opt,"-overburn"))
@@ -1199,7 +1523,7 @@ int main (int argc, char *argv[])
 		    exit(FATAL_START(EINVAL));
 	    }
 	    else if (!strncmp(opt,"-use-the-force-luke",19))
-	    { char *s=strchr (opt,'=');
+	    { char *s=strchr (opt,'='),*o;
 
 		if (s == NULL)	/* backward compatibility */
 		    no_tty_check = 1;
@@ -1207,15 +1531,45 @@ int main (int argc, char *argv[])
 		{   s++;
 		    if (strstr(s,"tty"))	no_tty_check = 1;
 		    if (strstr(s,"dummy"))	test_write   = 1;
-		    if (strstr(s,"dao"))	dvd_compat  += 256;
 		    if (strstr(s,"notray"))	no_reload    = 1;
 		    if (strstr(s,"moi"))
 		    {	quiet=-1; mkisofs_argv[mkisofs_argc++] = "-quiet";   }
+		    if ((o=strstr(s,"dao")))
+		    {	dvd_compat  += 256;
+			/*  vvvvvvvvvvv tracksize option takes precedence! */
+			if (dao_size==0 && (o[3]==':' || o[3]=='='))
+			{   dao_size=strtol(o+4,0,0);
+			    if (dao_size%16)
+				fprintf (stderr,":-( insane dao%c%d option\n",
+						o[3],dao_size),
+				exit(FATAL_START(EINVAL));
+			}
+		    }
+		    if ((o=strstr(s,"tracksize")))
+		    {	if (o[9]==':' || o[9]=='=')
+			{   dao_size=strtol(o+10,0,0);
+			    if (dao_size%16)
+				fprintf (stderr,":-( insane tracksize%c%d option\n",
+						o[9],dao_size),
+				exit(FATAL_START(EINVAL));
+			}
+		    }
+		    if ((o=strstr(s,"seek")) && next_session<0)
+		    {	if (o[4]==':' || o[4]=='=')
+			{   next_session=strtol(o+5,0,0);
+			    if (next_session%16)
+				fprintf (stderr,":-( insane seek%c%d option\n",
+						o[4],next_session),
+				exit(FATAL_START(EINVAL));
+			}
+		    }
 		}
 		continue;
 	    }
 	    else if (!strcmp(opt,"-dvd-video"))
-		poor_man = 1, dvd_compat++,	growisofs_argc++;
+	    {	if (poor_man<0) poor_man = 1;
+		dvd_compat++,	growisofs_argc++;
+	    }
 	    else if (!strcmp(opt,"-quiet"))
 		quiet++,			growisofs_argc++;
 	    else if (argv[i][1] == 'C' || !strncmp(opt,"-cdrecord-params",16))
@@ -1242,6 +1596,7 @@ int main (int argc, char *argv[])
 	    {	PRINT_VERSION (argv[0]);
 		printf ("  front-ending to %s: ",mkisofs_argv[0]);
 		fflush (stdout);
+		setuid(getuid());
 		execlp (mkisofs_argv[0],mkisofs_argv[0],"-version",NULL);
 		fprintf (stderr,"\n- %s: unable to execute %s: ",
 				argv[0],mkisofs_argv[0]),
@@ -1301,16 +1656,20 @@ int main (int argc, char *argv[])
 	     */
 	    out_device=setup_fds (in_device);
 
-	    *(long *)the_buffer=0;	/* redundant:-) */
-	    n=pread64 (in_fd,the_buffer,2048,VOLDESC_OFF*CD_BLOCK);
-	    if (n==0) errno=EIO;	/* end-of-file reached? */
-            if (n!=2048 && dev_found=='M')
-	        perror (":-( unable to pread64(2) primary volume descriptor"),
+	    *(long *)the_buffer=0;	/* redundant:-)	*/
+	    if (mmc_profile&0x10000)	/* blank media	*/
+		n=0, errno=EIO;
+	    else
+	    {	n=pread64 (in_fd,the_buffer,2048,VOLDESC_OFF*CD_BLOCK);
+		if (n==0) errno=EIO;	/* end-of-file reached? */
+	    }
+	    if (n!=2048 && dev_found=='M')
+		perror (":-( unable to pread64(2) primary volume descriptor"),
 		fprintf (stderr,"    you most likely want to use -Z option.\n"), 
 		exit (FATAL_START(errno));
 
 	    if (dev_found == 'M')
-            {	if (memcmp (the_buffer,"\1CD001",6))
+	    {	if (memcmp (the_buffer,"\1CD001",6))
 	            fprintf (stderr,":-( %s doesn't look like isofs...\n",
 		    		in_device), exit(FATAL_START(EMEDIUMTYPE));
 
@@ -1327,8 +1686,9 @@ int main (int argc, char *argv[])
 					    in_device,off),
 			    perror (NULL), exit(FATAL_START(errno));
 		    } else if (alleged_next_session!=next_session)
-			fprintf (stderr,"%s: -C argument is not present "
-					"or insane.\n",argv[0]),
+			fprintf (stderr,"%s: -C argument is %s.\n",
+					argv[0],alleged_next_session>=0?
+					"insane":"undefined"),
 			exit(FATAL_START(EINVAL));
 		}
 		else if (next_session > (0x200000-0x5000)) /* 4GB/2K-40MB/2K */
@@ -1338,10 +1698,18 @@ int main (int argc, char *argv[])
 
 		mkisofs_argv[mkisofs_argc++] = "-C";
 		mkisofs_argv[mkisofs_argc++] = C_parm;
-#ifdef PASS_STDIN_TO_MKISOFS
-		sprintf (M_parm,"/dev/fd/%d",0);
+#ifdef CANNOT_PASS_DEV_FD_N_TO_MKISOFS
+# ifdef PASS_STDIN_TO_MKISOFS
+		M_parm = "-";
+# else
+		M_parm = get_M_parm (in_fd,in_device);
+# endif
 #else
+# ifdef PASS_STDIN_TO_MKISOFS
+		M_parm = "/dev/fd/0";
+# else
 		sprintf (M_parm,"/dev/fd/%d",in_fd);
+# endif
 #endif
 		mkisofs_argv[mkisofs_argc++] = "-M";
 		mkisofs_argv[mkisofs_argc++] = M_parm;
@@ -1351,7 +1719,7 @@ int main (int argc, char *argv[])
 	    else
 	    {	if (!memcmp (the_buffer,"\1CD001",6))
 		    warn_for_isofs = 1;
-		next_session = 0;
+		if (next_session<0) next_session = 0;
 		continue;
 	    }
 	}
@@ -1382,13 +1750,23 @@ int main (int argc, char *argv[])
     assert (out_fd!=-1);
 
     /* never finalize disc at multi-sessioning DVDR recordings...	*/
-    if (next_session>0 &&
-	(mmc_profile==0x1B || mmc_profile==0x11))	dvd_compat=0;
-    /* ... except when filling the media up:-)				*/
-    if (next_session>0 && zero_image)			dvd_compat=1;
+    { int profile = mmc_profile&0xFFFF;
+	if (next_session>0 &&
+	    (profile==0x1B || profile==0x11))	dvd_compat=0;
+	/* ... except when filling the media up:-)			*/
+	if (next_session>0 && zero_image)	dvd_compat=1;
+    }
 
     if (warn_for_isofs)
-    {	if (no_tty_check || isatty (0))
+    { int fd=open("/dev/tty",O_RDONLY);
+
+	if (fd>=0)
+	{   if (isatty (fd)) warn_for_isofs |= 2;
+	    close (fd);
+	}
+	else if (isatty (0)) warn_for_isofs |= 2;
+
+	if (no_tty_check || (warn_for_isofs&2))
 	    fprintf (stderr,"WARNING: %s already carries isofs!\n",in_device),
 	    printf ("About to execute '");
 	else
@@ -1409,7 +1787,7 @@ int main (int argc, char *argv[])
     printf ("'\n");
     fflush (stdout);
 
-    if (warn_for_isofs && isatty(0) && !dry_run)
+    if ((warn_for_isofs&2) && !dry_run && !no_tty_check)
     {	fprintf (stderr,"Sleeping for 5 sec...\a"),	poll (NULL,0,1000);
 	fprintf (stderr,"\b\b\b\b\b\b\b\b4 sec...\a"),	poll (NULL,0,1000);
 	fprintf (stderr,"\b\b\b\b\b\b\b\b3 sec...\a"),	poll (NULL,0,1000);
@@ -1446,7 +1824,7 @@ int main (int argc, char *argv[])
 	}
 	while (1)
 	{   rpid = waitpid (pid,&rval,0);
-	    if (rpid == (pid_t)-1 || (rpid != pid && (errno=ECHILD)))
+	    if (rpid == (pid_t)-1 || (rpid != pid && (errno=ECHILD,1)))
 	    {	if (errno==EINTR)	continue;
 		else			perror (":-( waipid failed"),
 					exit(FATAL_START(errno));
@@ -1458,15 +1836,26 @@ int main (int argc, char *argv[])
 	    break;
 	}
 	if (errno)
-	    fprintf (stderr,":-( unable to unmount %s: ",in_device),
+	{   if (errno==EBUSY)
+		fprintf (stderr,":-( %s: unable to proceed with recording: ",
+			    in_device),
+#ifdef __hpux
+		fprintf (stderr,"device is mounted\n"),
+#else
+		fprintf (stderr,"unable to unmount\n"),
+#endif
+		exit (FATAL_START(errno));
+	    else
+		fprintf (stderr,":-( unable to umount %s: ",in_device);
 	    perror (""), exit (FATAL_START(errno));
+	}
     }
 
     if (poor_man)
     {	out_device=in_device;
 	if (!ioctl_device) ioctl_device=out_device;
 
-	switch (mmc_profile)
+	switch (mmc_profile&0xFFFF)
 	{   case 0x1A:	/* DVD+RW			*/
 	    case 0x1B:	/* DVD+R			*/
 	    case 0x11:	/* DVD-R Sequential		*/
